"""Defines fixtures that can be used to streamline tests and / or define dependencies"""
from collections import namedtuple
from random import randint

import hug
import pytest

Routers = namedtuple('Routers', ['http', 'local', 'cli'])


class TestAPI(hug.API):
    pass


@pytest.fixture
def hug_api():
    """Defines a dependency for and then includes a uniquely identified hug API for a single test case"""
    api = TestAPI('fake_api_{}'.format(randint(0, 1000000)))
    api.route = Routers(hug.routing.URLRouter().api(api),
                        hug.routing.LocalRouter().api(api),
                        hug.routing.CLIRouter().api(api))
    return api



"""tests/test_decorators.py.

Tests the decorators that power hugs core functionality

Copyright (C) 2016 Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import os
import sys
from unittest import mock

import falcon
import hug
import pytest
import requests
from falcon.testing import StartResponseMock, create_environ
from hug._async import coroutine

from .constants import BASE_DIRECTORY

api = hug.API(__name__)
module = sys.modules[__name__]

# Fix flake8 undefined names (F821)
__hug__ = __hug__  # noqa
__hug_wsgi__ = __hug_wsgi__  # noqa


def test_basic_call():
    """The most basic Happy-Path test for Hug APIs"""
    @hug.call()
    def hello_world():
        return "Hello World!"

    assert hello_world() == "Hello World!"
    assert hello_world.interface.http

    assert hug.test.get(api, '/hello_world').data == "Hello World!"
    assert hug.test.get(module, '/hello_world').data == "Hello World!"


def test_basic_call_on_method():
    """Test to ensure the most basic call still works if applied to a method"""
    class API(object):

        @hug.call()
        def hello_world(self=None):
            return "Hello World!"

    api_instance = API()
    assert api_instance.hello_world.interface.http
    assert api_instance.hello_world() == 'Hello World!'
    assert hug.test.get(api, '/hello_world').data == "Hello World!"

    class API(object):

        def hello_world(self):
            return "Hello World!"

    api_instance = API()

    @hug.call()
    def hello_world():
        return api_instance.hello_world()

    assert api_instance.hello_world() == 'Hello World!'
    assert hug.test.get(api, '/hello_world').data == "Hello World!"

    class API(object):

        def __init__(self):
            hug.call()(self.hello_world_method)

        def hello_world_method(self):
            return "Hello World!"

    api_instance = API()

    assert api_instance.hello_world_method() == 'Hello World!'
    assert hug.test.get(api, '/hello_world_method').data == "Hello World!"


def test_single_parameter():
    """Test that an api with a single parameter interacts as desired"""
    @hug.call()
    def echo(text):
        return text

    assert echo('Embrace') == 'Embrace'
    assert echo.interface.http
    with pytest.raises(TypeError):
        echo()

    assert hug.test.get(api, 'echo', text="Hello").data == "Hello"
    assert 'required' in hug.test.get(api, '/echo').data['errors']['text'].lower()


def test_on_invalid_transformer():
    """Test to ensure it is possible to transform data when data is invalid"""
    @hug.call(on_invalid=lambda data: 'error')
    def echo(text):
        return text
    assert hug.test.get(api, '/echo').data == 'error'


    def handle_error(data, request, response):
        return 'errored'

    @hug.call(on_invalid=handle_error)
    def echo2(text):
        return text
    assert hug.test.get(api, '/echo2').data == 'errored'


def test_on_invalid_format():
    """Test to ensure it's possible to change the format based on a validation error"""
    @hug.get(output_invalid=hug.output_format.json, output=hug.output_format.file)
    def echo(text):
        return text

    assert isinstance(hug.test.get(api, '/echo').data, dict)

    def smart_output_type(response, request):
        if response and request:
            return 'application/json'

    @hug.format.content_type(smart_output_type)
    def output_formatter(data, request, response):
        return hug.output_format.json((data, request and True, response and True))

    @hug.get(output_invalid=output_formatter, output=hug.output_format.file)
    def echo2(text):
        return text

    assert isinstance(hug.test.get(api, '/echo2').data, (list, tuple))


def test_smart_redirect_routing():
    """Test to ensure you can easily redirect to another method without an actual redirect"""
    @hug.get()
    def implementation_1():
        return 1

    @hug.get()
    def implementation_2():
        return 2

    @hug.get()
    def smart_route(implementation: int):
        if implementation == 1:
            return implementation_1
        elif implementation == 2:
            return implementation_2
        else:
            return "NOT IMPLEMENTED"

    assert hug.test.get(api, 'smart_route', implementation=1).data == 1
    assert hug.test.get(api, 'smart_route', implementation=2).data == 2
    assert hug.test.get(api, 'smart_route', implementation=3).data == "NOT IMPLEMENTED"


def test_custom_url():
    """Test to ensure that it's possible to have a route that differs from the function name"""
    @hug.call('/custom_route')
    def method_name():
        return 'works'

    assert hug.test.get(api, 'custom_route').data == 'works'


def test_api_auto_initiate():
    """Test to ensure that Hug automatically exposes a wsgi server method"""
    assert isinstance(__hug_wsgi__(create_environ('/non_existant'), StartResponseMock()), (list, tuple))


def test_parameters():
    """Tests to ensure that Hug can easily handle multiple parameters with multiple types"""
    @hug.call()
    def multiple_parameter_types(start, middle: hug.types.text, end: hug.types.number=5, **kwargs):
        return 'success'

    assert hug.test.get(api, 'multiple_parameter_types', start='start', middle='middle', end=7).data == 'success'
    assert hug.test.get(api, 'multiple_parameter_types', start='start', middle='middle').data == 'success'
    assert hug.test.get(api, 'multiple_parameter_types', start='start', middle='middle', other="yo").data == 'success'

    nan_test = hug.test.get(api, 'multiple_parameter_types', start='start', middle='middle', end='NAN').data
    assert 'Invalid' in nan_test['errors']['end']


def test_raise_on_invalid():
    """Test to ensure hug correctly respects a request to allow validations errors to pass through as exceptions"""
    @hug.get(raise_on_invalid=True)
    def my_handler(argument_1: int):
        return True

    with pytest.raises(Exception):
        hug.test.get(api, 'my_handler', argument_1='hi')

    assert hug.test.get(api, 'my_handler', argument_1=1)


def test_range_request():
    """Test to ensure that requesting a range works as expected"""
    @hug.get(output=hug.output_format.png_image)
    def image():
        return 'artwork/logo.png'

    assert hug.test.get(api, 'image', headers={'range': 'bytes=0-100'})
    assert hug.test.get(api, 'image', headers={'range': 'bytes=0--1'})

def test_parameters_override():
    """Test to ensure the parameters override is handled as expected"""
    @hug.get(parameters=('parameter1', 'parameter2'))
    def test_call(**kwargs):
        return kwargs

    assert hug.test.get(api, 'test_call', parameter1='one', parameter2='two').data == {'parameter1': 'one',
                                                                                       'parameter2': 'two'}


def test_parameter_injection():
    """Tests that hug correctly auto injects variables such as request and response"""
    @hug.call()
    def inject_request(request):
        return request and 'success'
    assert hug.test.get(api, 'inject_request').data == 'success'

    @hug.call()
    def inject_response(response):
        return response and 'success'
    assert hug.test.get(api, 'inject_response').data == 'success'

    @hug.call()
    def inject_both(request, response):
        return request and response and 'success'
    assert hug.test.get(api, 'inject_both').data == 'success'

    @hug.call()
    def wont_appear_in_kwargs(**kwargs):
        return 'request' not in kwargs and 'response' not in kwargs and 'success'
    assert hug.test.get(api, 'wont_appear_in_kwargs').data == 'success'


def test_method_routing():
    """Test that all hugs HTTP routers correctly route methods to the correct handler"""
    @hug.get()
    def method_get():
        return 'GET'

    @hug.post()
    def method_post():
        return 'POST'

    @hug.connect()
    def method_connect():
        return 'CONNECT'

    @hug.delete()
    def method_delete():
        return 'DELETE'

    @hug.options()
    def method_options():
        return 'OPTIONS'

    @hug.put()
    def method_put():
        return 'PUT'

    @hug.trace()
    def method_trace():
        return 'TRACE'

    assert hug.test.get(api, 'method_get').data == 'GET'
    assert hug.test.post(api, 'method_post').data == 'POST'
    assert hug.test.connect(api, 'method_connect').data == 'CONNECT'
    assert hug.test.delete(api, 'method_delete').data == 'DELETE'
    assert hug.test.options(api, 'method_options').data == 'OPTIONS'
    assert hug.test.put(api, 'method_put').data == 'PUT'
    assert hug.test.trace(api, 'method_trace').data == 'TRACE'

    @hug.call(accept=('GET', 'POST'))
    def accepts_get_and_post():
        return 'success'

    assert hug.test.get(api, 'accepts_get_and_post').data == 'success'
    assert hug.test.post(api, 'accepts_get_and_post').data == 'success'
    assert 'method not allowed' in hug.test.trace(api, 'accepts_get_and_post').status.lower()


def test_not_found():
    """Test to ensure the not_found decorator correctly routes 404s to the correct handler"""
    @hug.not_found()
    def not_found_handler():
        return "Not Found"

    result = hug.test.get(api, '/does_not_exist/yet')
    assert result.data == "Not Found"
    assert result.status == falcon.HTTP_NOT_FOUND

    @hug.not_found(versions=10)  # noqa
    def not_found_handler(response):
        response.status = falcon.HTTP_OK
        return {'look': 'elsewhere'}

    result = hug.test.get(api, '/v10/does_not_exist/yet')
    assert result.data == {'look': 'elsewhere'}
    assert result.status == falcon.HTTP_OK

    result = hug.test.get(api, '/does_not_exist/yet')
    assert result.data == "Not Found"
    assert result.status == falcon.HTTP_NOT_FOUND

    del api.http._not_found_handlers


def test_not_found_with_extended_api():
    """Test to ensure the not_found decorator works correctly when the API is extended"""
    @hug.extend_api()
    def extend_with():
        import tests.module_fake
        return (tests.module_fake, )

    assert hug.test.get(api, '/does_not_exist/yet').data is True

def test_versioning():
    """Ensure that Hug correctly routes API functions based on version"""
    @hug.get('/echo')
    def echo(text):
        return "Not Implemented"

    @hug.get('/echo', versions=1)  # noqa
    def echo(text):
        return text

    @hug.get('/echo', versions=range(2, 4))  # noqa
    def echo(text):
        return "Echo: {text}".format(**locals())

    @hug.get('/echo', versions=7)  # noqa
    def echo(text, api_version):
        return api_version

    @hug.get('/echo', versions=False)  # noqa
    def echo(text):
        return "No Versions"

    assert hug.test.get(api, 'v1/echo', text="hi").data == 'hi'
    assert hug.test.get(api, 'v2/echo', text="hi").data == "Echo: hi"
    assert hug.test.get(api, 'v3/echo', text="hi").data == "Echo: hi"
    assert hug.test.get(api, 'echo', text="hi", api_version=3).data == "Echo: hi"
    assert hug.test.get(api, 'echo', text="hi", headers={'X-API-VERSION': '3'}).data == "Echo: hi"
    assert hug.test.get(api, 'v4/echo', text="hi").data == "Not Implemented"
    assert hug.test.get(api, 'v7/echo', text="hi").data == 7
    assert hug.test.get(api, 'echo', text="hi").data == "No Versions"
    assert hug.test.get(api, 'echo', text="hi", api_version=3, body={'api_vertion': 4}).data == "Echo: hi"

    with pytest.raises(ValueError):
        hug.test.get(api, 'v4/echo', text="hi", api_version=3)


def test_multiple_version_injection():
    """Test to ensure that the version injected sticks when calling other functions within an API"""
    @hug.get(versions=(1, 2, None))
    def my_api_function(hug_api_version):
        return hug_api_version

    assert hug.test.get(api, 'v1/my_api_function').data == 1
    assert hug.test.get(api, 'v2/my_api_function').data == 2
    assert hug.test.get(api, 'v3/my_api_function').data == 3

    @hug.get(versions=(None, 1))
    @hug.local(version=1)
    def call_other_function(hug_current_api):
        return hug_current_api.my_api_function()

    assert hug.test.get(api, 'v1/call_other_function').data == 1
    assert call_other_function() == 1

    @hug.get(versions=1)
    @hug.local(version=1)
    def one_more_level_of_indirection(hug_current_api):
        return hug_current_api.call_other_function()

    assert hug.test.get(api, 'v1/one_more_level_of_indirection').data == 1
    assert one_more_level_of_indirection() == 1


def test_json_auto_convert():
    """Test to ensure all types of data correctly auto convert into json"""
    @hug.get('/test_json')
    def test_json(text):
        return text
    assert hug.test.get(api, 'test_json', body={'text': 'value'}).data == "value"

    @hug.get('/test_json_body')
    def test_json_body(body):
        return body
    assert hug.test.get(api, 'test_json_body', body=['value1', 'value2']).data == ['value1', 'value2']

    @hug.get(parse_body=False)
    def test_json_body_stream_only(body=None):
        return body
    assert hug.test.get(api, 'test_json_body_stream_only', body=['value1', 'value2']).data is None


def test_error_handling():
    """Test to ensure Hug correctly handles Falcon errors that are thrown during processing"""
    @hug.get()
    def test_error():
        raise falcon.HTTPInternalServerError('Failed', 'For Science!')

    response = hug.test.get(api, 'test_error')
    assert 'errors' in response.data
    assert response.data['errors']['Failed'] == 'For Science!'


def test_error_handling_builtin_exception():
    """Test to ensure built in exception types errors are handled as expected"""
    def raise_error(value):
        raise KeyError('Invalid value')

    @hug.get()
    def test_error(data: raise_error):
        return True

    response = hug.test.get(api, 'test_error', data=1)
    assert 'errors' in response.data
    assert response.data['errors']['data'] == 'Invalid value'


def test_error_handling_custom():
    """Test to ensure custom exceptions work as expected"""
    class Error(Exception):

        def __str__(self):
            return 'Error'

    def raise_error(value):
        raise Error()

    @hug.get()
    def test_error(data: raise_error):
        return True

    response = hug.test.get(api, 'test_error', data=1)
    assert 'errors' in response.data
    assert response.data['errors']['data'] == 'Error'


def test_return_modifer():
    """Ensures you can modify the output of a HUG API using -> annotation"""
    @hug.get()
    def hello() -> lambda data: "Hello {0}!".format(data):
        return "world"

    assert hug.test.get(api, 'hello').data == "Hello world!"
    assert hello() == 'world'

    @hug.get(transform=lambda data: "Goodbye {0}!".format(data))
    def hello() -> lambda data: "Hello {0}!".format(data):
        return "world"
    assert hug.test.get(api, 'hello').data == "Goodbye world!"
    assert hello() == 'world'

    @hug.get()
    def hello() -> str:
        return "world"
    assert hug.test.get(api, 'hello').data == "world"
    assert hello() == 'world'

    @hug.get(transform=False)
    def hello() -> lambda data: "Hello {0}!".format(data):
        return "world"

    assert hug.test.get(api, 'hello').data == "world"
    assert hello() == 'world'

    def transform_with_request_data(data, request, response):
        return (data, request and True, response and True)

    @hug.get(transform=transform_with_request_data)
    def hello():
        return "world"

    response = hug.test.get(api, 'hello')
    assert response.data == ['world', True, True]


def test_custom_deserializer_support():
    """Ensure that custom desirializers work as expected"""
    class CustomDeserializer(object):
        def from_string(self, string):
            return 'custom {}'.format(string)

    @hug.get()
    def test_custom_deserializer(text: CustomDeserializer()):
        return text

    assert hug.test.get(api, 'test_custom_deserializer', text='world').data == 'custom world'


def test_marshmallow_support():
    """Ensure that you can use Marshmallow style objects to control input and output validation and transformation"""
    class MarshmallowStyleObject(object):
        def dump(self, item):
            return 'Dump Success'

        def load(self, item):
            return ('Load Success', None)

        def loads(self, item):
            return self.load(item)

    schema = MarshmallowStyleObject()

    @hug.get()
    def test_marshmallow_style() -> schema:
        return 'world'

    assert hug.test.get(api, 'test_marshmallow_style').data == "Dump Success"
    assert test_marshmallow_style() == 'world'


    @hug.get()
    def test_marshmallow_input(item: schema):
        return item

    assert hug.test.get(api, 'test_marshmallow_input', item='bacon').data == "Load Success"
    assert test_marshmallow_style() == 'world'

    class MarshmallowStyleObjectWithError(object):
        def dump(self, item):
            return 'Dump Success'

        def load(self, item):
            return ('Load Success', {'type': 'invalid'})

        def loads(self, item):
            return self.load(item)

    schema = MarshmallowStyleObjectWithError()

    @hug.get()
    def test_marshmallow_input2(item: schema):
        return item

    assert hug.test.get(api, 'test_marshmallow_input2', item='bacon').data == {'errors': {'item': {'type': 'invalid'}}}

    class MarshmallowStyleField(object):
        def deserialize(self, value):
            return str(value)

    @hug.get()
    def test_marshmallow_input_field(item: MarshmallowStyleField()):
        return item

    assert hug.test.get(api, 'test_marshmallow_input_field', item='bacon').data == 'bacon'


def test_stream_return():
    """Test to ensure that its valid for a hug API endpoint to return a stream"""
    @hug.get(output=hug.output_format.text)
    def test():
        return open(os.path.join(BASE_DIRECTORY, 'README.md'), 'rb')

    assert 'hug' in hug.test.get(api, 'test').data


def test_smart_outputter():
    """Test to ensure that the output formatter can accept request and response arguments"""
    def smart_output_type(response, request):
        if response and request:
            return 'application/json'

    @hug.format.content_type(smart_output_type)
    def output_formatter(data, request, response):
        return hug.output_format.json((data, request and True, response and True))

    @hug.get(output=output_formatter)
    def test():
        return True

    assert hug.test.get(api, 'test').data == [True, True, True]


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_output_format():
    """Test to ensure it's possible to quickly change the default hug output format"""
    old_formatter = api.http.output_format

    @hug.default_output_format()
    def augmented(data):
        return hug.output_format.json(['Augmented', data])

    @hug.get(suffixes=('.js', '/js'), prefixes='/text')
    def hello():
        return "world"

    assert hug.test.get(api, 'hello').data == ['Augmented', 'world']
    assert hug.test.get(api, 'hello.js').data == ['Augmented', 'world']
    assert hug.test.get(api, 'hello/js').data == ['Augmented', 'world']
    assert hug.test.get(api, 'text/hello').data == ['Augmented', 'world']

    @hug.default_output_format()
    def jsonify(data):
        return hug.output_format.json(data)


    api.http.output_format = hug.output_format.text

    @hug.get()
    def my_method():
        return {'Should': 'work'}

    assert hug.test.get(api, 'my_method').data == "{'Should': 'work'}"
    api.http.output_format = old_formatter


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_input_format():
    """Test to ensure it's possible to quickly change the default hug output format"""
    old_format = api.http.input_format('application/json')
    api.http.set_input_format('application/json', lambda a: {'no': 'relation'})

    @hug.get()
    def hello(body):
        return body

    assert hug.test.get(api, 'hello', body={'should': 'work'}).data == {'no': 'relation'}

    @hug.get()
    def hello2(body):
        return body

    assert not hug.test.get(api, 'hello2').data

    api.http.set_input_format('application/json', old_format)


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_specific_input_format():
    """Test to ensure the input formatter can be specified"""
    @hug.get(inputs={'application/json': lambda a: 'formatted'})
    def hello(body):
        return body

    assert hug.test.get(api, 'hello', body={'should': 'work'}).data == 'formatted'


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_content_type_with_parameter():
    """Test a Content-Type with parameter as `application/json charset=UTF-8`
    as described in https://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7"""
    @hug.get()
    def demo(body):
        return body

    assert hug.test.get(api, 'demo', body={}, headers={'content-type': 'application/json'}).data == {}
    assert hug.test.get(api, 'demo', body={}, headers={'content-type': 'application/json; charset=UTF-8'}).data == {}


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_middleware():
    """Test to ensure the basic concept of a middleware works as expected"""
    @hug.request_middleware()
    def proccess_data(request, response):
        request.env['SERVER_NAME'] = 'Bacon'

    @hug.response_middleware()
    def proccess_data2(request, response, resource):
        response.set_header('Bacon', 'Yumm')

    @hug.reqresp_middleware()
    def process_data3(request):
        request.env['MEET'] = 'Ham'
        response, resource = yield request
        response.set_header('Ham', 'Buu!!')
        yield response

    @hug.get()
    def hello(request):
        return [
            request.env['SERVER_NAME'],
            request.env['MEET']
        ]

    result = hug.test.get(api, 'hello')
    assert result.data == ['Bacon', 'Ham']
    assert result.headers_dict['Bacon'] == 'Yumm'
    assert result.headers_dict['Ham'] == 'Buu!!'


def test_requires():
    """Test to ensure only if requirements successfully keep calls from happening"""
    def user_is_not_tim(request, response, **kwargs):
        if request.headers.get('USER', '') != 'Tim':
            return True
        return 'Unauthorized'

    @hug.get(requires=user_is_not_tim)
    def hello(request):
        return 'Hi!'

    assert hug.test.get(api, 'hello').data == 'Hi!'
    assert hug.test.get(api, 'hello', headers={'USER': 'Tim'}).data == 'Unauthorized'


def test_extending_api():
    """Test to ensure it's possible to extend the current API from an external file"""
    @hug.extend_api('/fake')
    def extend_with():
        import tests.module_fake
        return (tests.module_fake, )

    assert hug.test.get(api, 'fake/made_up_api').data


def test_extending_api_simple():
    """Test to ensure it's possible to extend the current API from an external file with just one API endpoint"""
    @hug.extend_api('/fake_simple')
    def extend_with():
        import tests.module_fake_simple
        return (tests.module_fake_simple, )

    assert hug.test.get(api, 'fake_simple/made_up_hello').data == 'hello'


def test_extending_api_with_exception_handler():
    """Test to ensure it's possible to extend the current API from an external file"""

    from tests.module_fake_simple import FakeSimpleException

    @hug.exception(FakeSimpleException)
    def handle_exception(exception):
        return 'it works!'

    @hug.extend_api('/fake_simple')
    def extend_with():
        import tests.module_fake_simple
        return (tests.module_fake_simple, )

    assert hug.test.get(api, '/fake_simple/exception').data == 'it works!'


def test_extending_api_with_base_url():
    """Test to ensure it's possible to extend the current API with a specified base URL"""
    @hug.extend_api('/fake', base_url='/api')
    def extend_with():
        import tests.module_fake
        return (tests.module_fake, )

    assert hug.test.get(api, '/api/v1/fake/made_up_api').data


def test_extending_api_with_same_path_under_different_base_url():
    """Test to ensure it's possible to extend the current API with the same path under a different base URL"""
    @hug.get()
    def made_up_hello():
        return 'hi'

    @hug.extend_api(base_url='/api')
    def extend_with():
        import tests.module_fake_simple
        return (tests.module_fake_simple, )

    assert hug.test.get(api, '/made_up_hello').data == 'hi'
    assert hug.test.get(api, '/api/made_up_hello').data == 'hello'


def test_cli():
    """Test to ensure the CLI wrapper works as intended"""
    @hug.cli('command', '1.0.0', output=str)
    def cli_command(name: str, value: int):
        return (name, value)

    assert cli_command('Testing', 1) == ('Testing', 1)
    assert hug.test.cli(cli_command, "Bob", 5) == ("Bob", 5)


def test_cli_requires():
    """Test to ensure your can add requirements to a CLI"""
    def requires_fail(**kwargs):
        return {'requirements': 'not met'}

    @hug.cli(output=str, requires=requires_fail)
    def cli_command(name: str, value: int):
        return (name, value)

    assert cli_command('Testing', 1) == ('Testing', 1)
    assert hug.test.cli(cli_command, 'Testing', 1) == {'requirements': 'not met'}


def test_cli_validation():
    """Test to ensure your can add custom validation to a CLI"""
    def contains_either(fields):
        if not fields.get('name', '') and not fields.get('value', 0):
            return {'name': 'must be defined', 'value': 'must be defined'}

    @hug.cli(output=str, validate=contains_either)
    def cli_command(name: str="", value: int=0):
        return (name, value)

    assert cli_command('Testing', 1) == ('Testing', 1)
    assert hug.test.cli(cli_command) == {'name': 'must be defined', 'value': 'must be defined'}
    assert hug.test.cli(cli_command, name='Testing') == ('Testing', 0)


def test_cli_with_defaults():
    """Test to ensure CLIs work correctly with default values"""
    @hug.cli()
    def happy(name: str, age: int, birthday: bool=False):
        if birthday:
            return "Happy {age} birthday {name}!".format(**locals())
        else:
            return "{name} is {age} years old".format(**locals())

    assert happy('Hug', 1) == "Hug is 1 years old"
    assert happy('Hug', 1, True) == "Happy 1 birthday Hug!"
    assert hug.test.cli(happy, "Bob", 5) == "Bob is 5 years old"
    assert hug.test.cli(happy, "Bob", 5, birthday=True) == "Happy 5 birthday Bob!"


def test_cli_with_hug_types():
    """Test to ensure CLIs work as expected when using hug types"""
    @hug.cli()
    def happy(name: hug.types.text, age: hug.types.number, birthday: hug.types.boolean=False):
        if birthday:
            return "Happy {age} birthday {name}!".format(**locals())
        else:
            return "{name} is {age} years old".format(**locals())

    assert happy('Hug', 1) == "Hug is 1 years old"
    assert happy('Hug', 1, True) == "Happy 1 birthday Hug!"
    assert hug.test.cli(happy, "Bob", 5) == "Bob is 5 years old"
    assert hug.test.cli(happy, "Bob", 5, birthday=True) == "Happy 5 birthday Bob!"

    @hug.cli()
    def succeed(success: hug.types.smart_boolean=False):
        if success:
            return 'Yes!'
        else:
            return 'No :('

    assert hug.test.cli(succeed) == 'No :('
    assert hug.test.cli(succeed, success=True) == 'Yes!'
    assert 'succeed' in str(__hug__.cli)

    @hug.cli()
    def succeed(success: hug.types.smart_boolean=True):
        if success:
            return 'Yes!'
        else:
            return 'No :('

    assert hug.test.cli(succeed) == 'Yes!'
    assert hug.test.cli(succeed, success='false') == 'No :('

    @hug.cli()
    def all_the(types: hug.types.multiple=[]):
        return types or ['nothing_here']

    assert hug.test.cli(all_the) == ['nothing_here']
    assert hug.test.cli(all_the, types=('one', 'two', 'three')) == ['one', 'two', 'three']

    @hug.cli()
    def all_the(types: hug.types.multiple):
        return types or ['nothing_here']

    assert hug.test.cli(all_the) == ['nothing_here']
    assert hug.test.cli(all_the, 'one', 'two', 'three') == ['one', 'two', 'three']

    @hug.cli()
    def one_of(value: hug.types.one_of(['one', 'two'])='one'):
        return value

    assert hug.test.cli(one_of, value='one') == 'one'
    assert hug.test.cli(one_of, value='two') == 'two'


def test_cli_with_conflicting_short_options():
    """Test to ensure that it's possible to expose a CLI with the same first few letters in option"""
    @hug.cli()
    def test(abe1="Value", abe2="Value2", helper=None):
        return (abe1, abe2)

    assert test() == ('Value', 'Value2')
    assert test('hi', 'there') == ('hi', 'there')
    assert hug.test.cli(test) == ('Value', 'Value2')
    assert hug.test.cli(test, abe1='hi', abe2='there') == ('hi', 'there')


def test_cli_with_directives():
    """Test to ensure it's possible to use directives with hug CLIs"""
    @hug.cli()
    @hug.local()
    def test(hug_timer):
        return float(hug_timer)

    assert isinstance(test(), float)
    assert test(hug_timer=4) == 4
    assert isinstance(hug.test.cli(test), float)


def test_cli_with_named_directives():
    """Test to ensure you can pass named directives into the cli"""
    @hug.cli()
    @hug.local()
    def test(timer: hug.directives.Timer):
        return float(timer)

    assert isinstance(test(), float)
    assert test(timer=4) == 4
    assert isinstance(hug.test.cli(test), float)


def test_cli_with_output_transform():
    """Test to ensure it's possible to use output transforms with hug CLIs"""
    @hug.cli()
    def test() -> int:
        return '5'

    assert isinstance(test(), str)
    assert isinstance(hug.test.cli(test), int)


    @hug.cli(transform=int)
    def test():
        return '5'

    assert isinstance(test(), str)
    assert isinstance(hug.test.cli(test), int)


def test_cli_with_short_short_options():
    """Test to ensure that it's possible to expose a CLI with 2 very short and similar options"""
    @hug.cli()
    def test(a1="Value", a2="Value2"):
        return (a1, a2)

    assert test() == ('Value', 'Value2')
    assert test('hi', 'there') == ('hi', 'there')
    assert hug.test.cli(test) == ('Value', 'Value2')
    assert hug.test.cli(test, a1='hi', a2='there') == ('hi', 'there')


def test_cli_file_return():
    """Test to ensure that its possible to return a file stream from a CLI"""
    @hug.cli()
    def test():
        return open(os.path.join(BASE_DIRECTORY, 'README.md'), 'rb')

    assert 'hug' in hug.test.cli(test)


def test_local_type_annotation():
    """Test to ensure local type annotation works as expected"""
    @hug.local(raise_on_invalid=True)
    def test(number: int):
        return number

    assert test(3) == 3
    with pytest.raises(Exception):
        test('h')

    @hug.local(raise_on_invalid=False)
    def test(number: int):
        return number

    assert test('h')['errors']

    @hug.local(raise_on_invalid=False, validate=False)
    def test(number: int):
        return number

    assert test('h') == 'h'


def test_local_transform():
    """Test to ensure local type annotation works as expected"""
    @hug.local(transform=str)
    def test(number: int):
        return number

    assert test(3) == '3'


def test_local_on_invalid():
    """Test to ensure local type annotation works as expected"""
    @hug.local(on_invalid=str)
    def test(number: int):
        return number

    assert isinstance(test('h'), str)


def test_local_requires():
    """Test to ensure only if requirements successfully keep calls from happening locally"""
    global_state = False

    def requirement(**kwargs):
        return global_state and 'Unauthorized'

    @hug.local(requires=requirement)
    def hello():
        return 'Hi!'

    assert hello() == 'Hi!'
    global_state = True
    assert hello() == 'Unauthorized'


def test_static_file_support():
    """Test to ensure static file routing works as expected"""
    @hug.static('/static')
    def my_static_dirs():
        return (BASE_DIRECTORY, )

    assert 'hug' in hug.test.get(api, '/static/README.md').data
    assert 'Index' in hug.test.get(api, '/static/tests/data').data
    assert '404' in hug.test.get(api, '/static/NOT_IN_EXISTANCE.md').status


def test_static_jailed():
    """Test to ensure we can't serve from outside static dir"""
    @hug.static('/static')
    def my_static_dirs():
        return ['tests']
    assert '404' in hug.test.get(api, '/static/../README.md').status


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_sink_support():
    """Test to ensure sink URL routers work as expected"""
    @hug.sink('/all')
    def my_sink(request):
        return request.path.replace('/all', '')

    assert hug.test.get(api, '/all/the/things').data == '/the/things'

@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_sink_support_with_base_url():
    """Test to ensure sink URL routers work when the API is extended with a specified base URL"""
    @hug.extend_api('/fake', base_url='/api')
    def extend_with():
        import tests.module_fake
        return (tests.module_fake, )

    assert hug.test.get(api, '/api/fake/all/the/things').data == '/the/things'

def test_cli_with_string_annotation():
    """Test to ensure CLI's work correctly with string annotations"""
    @hug.cli()
    def test(value_1: 'The first value', value_2: 'The second value'=None):
        return True

    assert hug.test.cli(test, True)


def test_cli_with_args():
    """Test to ensure CLI's work correctly when taking args"""
    @hug.cli()
    def test(*values):
        return values

    assert test(1, 2, 3) == (1, 2, 3)
    assert hug.test.cli(test, 1, 2, 3) == ('1', '2', '3')


def test_cli_using_method():
    """Test to ensure that attaching a cli to a class method works as expected"""
    class API(object):

        def __init__(self):
            hug.cli()(self.hello_world_method)

        def hello_world_method(self):
            variable = 'Hello World!'
            return variable

    api_instance = API()
    assert api_instance.hello_world_method() == 'Hello World!'
    assert hug.test.cli(api_instance.hello_world_method) == 'Hello World!'
    assert hug.test.cli(api_instance.hello_world_method, collect_output=False) is None


def test_cli_with_nested_variables():
    """Test to ensure that a cli containing multiple nested variables works correctly"""
    @hug.cli()
    def test(value_1=None, value_2=None):
        return 'Hi!'

    assert hug.test.cli(test) == 'Hi!'


def test_cli_with_exception():
    """Test to ensure that a cli with an exception is correctly handled"""
    @hug.cli()
    def test():
        raise ValueError()
        return 'Hi!'

    assert hug.test.cli(test) != 'Hi!'


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_wraps():
    """Test to ensure you can safely apply decorators to hug endpoints by using @hug.wraps"""
    def my_decorator(function):
        @hug.wraps(function)
        def decorated(*args, **kwargs):
            kwargs['name'] = 'Timothy'
            return function(*args, **kwargs)
        return decorated

    @hug.get()
    @my_decorator
    def what_is_my_name(hug_timer=None, name="Sam"):
        return {'name': name, 'took': hug_timer}

    result = hug.test.get(api, 'what_is_my_name').data
    assert result['name'] == 'Timothy'
    assert result['took']

    def my_second_decorator(function):
        @hug.wraps(function)
        def decorated(*args, **kwargs):
            kwargs['name'] = "Not telling"
            return function(*args, **kwargs)
        return decorated

    @hug.get()
    @my_decorator
    @my_second_decorator
    def what_is_my_name2(hug_timer=None, name="Sam"):
        return {'name': name, 'took': hug_timer}

    result = hug.test.get(api, 'what_is_my_name2').data
    assert result['name'] == "Not telling"
    assert result['took']

    def my_decorator_with_request(function):
        @hug.wraps(function)
        def decorated(request, *args, **kwargs):
            kwargs['has_request'] = bool(request)
            return function(*args, **kwargs)
        return decorated

    @hug.get()
    @my_decorator_with_request
    def do_you_have_request(has_request=False):
        return has_request

    assert hug.test.get(api, 'do_you_have_request').data


def test_cli_with_empty_return():
    """Test to ensure that if you return None no data will be added to sys.stdout"""
    @hug.cli()
    def test_empty_return():
        pass

    assert not hug.test.cli(test_empty_return)


def test_cli_with_multiple_ints():
    """Test to ensure multiple ints work with CLI"""
    @hug.cli()
    def test_multiple_cli(ints: hug.types.comma_separated_list):
        return ints

    assert hug.test.cli(test_multiple_cli, ints='1,2,3') == ['1', '2', '3']


    class ListOfInts(hug.types.Multiple):
        """Only accept a list of numbers."""

        def __call__(self, value):
            value = super().__call__(value)
            return [int(number) for number in value]

    @hug.cli()
    def test_multiple_cli(ints: ListOfInts()=[]):
        return ints

    assert hug.test.cli(test_multiple_cli, ints=['1', '2', '3']) == [1, 2, 3]


def test_startup():
    """Test to ensure hug startup decorators work as expected"""
    @hug.startup()
    def happens_on_startup(api):
        pass

    @hug.startup()
    @coroutine
    def async_happens_on_startup(api):
        pass

    assert happens_on_startup in api.startup_handlers
    assert async_happens_on_startup in api.startup_handlers


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_adding_headers():
    """Test to ensure it is possible to inject response headers based on only the URL route"""
    @hug.get(response_headers={'name': 'Timothy'})
    def endpoint():
        return ''

    result = hug.test.get(api, 'endpoint')
    assert result.data == ''
    assert result.headers_dict['name'] == 'Timothy'


def test_on_demand_404(hug_api):
    """Test to ensure it's possible to route to a 404 response on demand"""
    @hug_api.route.http.get()
    def my_endpoint(hug_api):
        return hug_api.http.not_found

    assert '404' in hug.test.get(hug_api, 'my_endpoint').status


    @hug_api.route.http.get()
    def my_endpoint2(hug_api):
        raise hug.HTTPNotFound()

    assert '404' in hug.test.get(hug_api, 'my_endpoint2').status

    @hug_api.route.http.get()
    def my_endpoint3(hug_api):
        """Test to ensure base 404 handler works as expected"""
        del hug_api.http._not_found
        return hug_api.http.not_found

    assert '404' in hug.test.get(hug_api, 'my_endpoint3').status


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_exceptions():
    """Test to ensure hug's exception handling decorator works as expected"""
    @hug.get()
    def endpoint():
        raise ValueError('hi')

    with pytest.raises(ValueError):
        hug.test.get(api, 'endpoint')

    @hug.exception()
    def handle_exception(exception):
        return 'it worked'

    assert hug.test.get(api, 'endpoint').data == 'it worked'

    @hug.exception(ValueError)  # noqa
    def handle_exception(exception):
        return 'more explicit handler also worked'

    assert hug.test.get(api, 'endpoint').data == 'more explicit handler also worked'


@pytest.mark.skipif(sys.platform == 'win32', reason='Currently failing on Windows build')
def test_validate():
    """Test to ensure hug's secondary validation mechanism works as expected"""
    def contains_either(fields):
        if not 'one' in fields and not 'two' in fields:
            return {'one': 'must be defined', 'two': 'must be defined'}

    @hug.get(validate=contains_either)
    def my_endpoint(one=None, two=None):
        return True


    assert hug.test.get(api, 'my_endpoint', one=True).data
    assert hug.test.get(api, 'my_endpoint', two=True).data
    assert hug.test.get(api, 'my_endpoint').status
    assert hug.test.get(api, 'my_endpoint').data == {'errors': {'one': 'must be defined', 'two': 'must be defined'}}


def test_cli_api(capsys):
    """Ensure that the overall CLI Interface API works as expected"""
    @hug.cli()
    def my_cli_command():
        print("Success!")

    with mock.patch('sys.argv', ['/bin/command', 'my_cli_command']):
        __hug__.cli()
        out, err = capsys.readouterr()
        assert "Success!" in out

    with mock.patch('sys.argv', []):
        with pytest.raises(SystemExit):
            __hug__.cli()


def test_cli_api_return():
    """Ensure returning from a CLI API works as expected"""
    @hug.cli()
    def my_cli_command():
        return "Success!"

    my_cli_command.interface.cli()


def test_urlencoded():
    """Ensure that urlencoded input format works as intended"""
    @hug.post()
    def test_url_encoded_post(**kwargs):
        return kwargs

    test_data = b'foo=baz&foo=bar&name=John+Doe'
    assert hug.test.post(api, 'test_url_encoded_post', body=test_data, headers={'content-type': 'application/x-www-form-urlencoded'}).data == {'name': 'John Doe', 'foo': ['baz', 'bar']}


def test_multipart():
    """Ensure that multipart input format works as intended"""
    @hug.post()
    def test_multipart_post(**kwargs):
        return kwargs

    with open(os.path.join(BASE_DIRECTORY, 'artwork', 'logo.png'),'rb') as logo:
        prepared_request = requests.Request('POST', 'http://localhost/', files={'logo': logo}).prepare()
        logo.seek(0)
        output = json.loads(hug.defaults.output_format({'logo': logo.read()}).decode('utf8'))
        assert hug.test.post(api, 'test_multipart_post',  body=prepared_request.body,
                             headers=prepared_request.headers).data == output


def test_json_null(hug_api):
    """Test to ensure passing in null within JSON will be seen as None and not allowed by text values"""
    @hug_api.route.http.post()
    def test_naive(argument_1):
        return argument_1

    assert hug.test.post(hug_api, 'test_naive', body='{"argument_1": null}',
                         headers={'content-type': 'application/json'}).data == None


    @hug_api.route.http.post()
    def test_text_type(argument_1: hug.types.text):
        return argument_1


    assert 'errors' in hug.test.post(hug_api, 'test_text_type', body='{"argument_1": null}',
                                    headers={'content-type': 'application/json'}).data


def test_json_self_key(hug_api):
    """Test to ensure passing in a json with a key named 'self' works as expected"""
    @hug_api.route.http.post()
    def test_self_post(body):
        return body

    assert hug.test.post(hug_api, 'test_self_post', body='{"self": "this"}',
                         headers={'content-type': 'application/json'}).data == {"self": "this"}


def test_204_with_no_body(hug_api):
    """Test to ensure returning no body on a 204 statused endpoint works without issue"""
    @hug_api.route.http.delete()
    def test_route(response):
        response.status = hug.HTTP_204
        return

    assert '204' in hug.test.delete(hug_api, 'test_route').status


def test_output_format_inclusion(hug_api):
    """Test to ensure output format can live in one api but apply to the other"""
    @hug.get()
    def my_endpoint():
        return 'hello'

    @hug.default_output_format(api=hug_api)
    def mutated_json(data):
        return hug.output_format.json({'mutated': data})

    hug_api.extend(api, '')

    assert hug.test.get(hug_api, 'my_endpoint').data == {'mutated': 'hello'}


def test_api_pass_along(hug_api):
    """Test to ensure the correct API instance is passed along using API directive"""
    @hug.get()
    def takes_api(hug_api):
        return hug_api.__name__

    hug_api.__name__ = "Test API"
    hug_api.extend(api, '')
    assert hug.test.get(hug_api, 'takes_api').data == hug_api.__name__


def test_exception_excludes(hug_api):
    """Test to ensure it's possible to add excludes to exception routers"""
    class MyValueError(ValueError):
        pass

    class MySecondValueError(ValueError):
        pass

    @hug.exception(Exception, exclude=MySecondValueError, api=hug_api)
    def base_exception_handler(exception):
        return 'base exception handler'

    @hug.exception(ValueError, exclude=(MyValueError, MySecondValueError), api=hug_api)
    def base_exception_handler(exception):
        return 'special exception handler'

    @hug.get(api=hug_api)
    def my_handler():
        raise MyValueError()

    @hug.get(api=hug_api)
    def fall_through_handler():
        raise ValueError('reason')

    @hug.get(api=hug_api)
    def full_through_to_raise():
        raise MySecondValueError()

    assert hug.test.get(hug_api, 'my_handler').data == 'base exception handler'
    assert hug.test.get(hug_api, 'fall_through_handler').data == 'special exception handler'
    with pytest.raises(MySecondValueError):
        assert hug.test.get(hug_api, 'full_through_to_raise').data


def test_cli_kwargs(hug_api):
    """Test to ensure cli commands can correctly handle **kwargs"""
    @hug.cli(api=hug_api)
    def takes_all_the_things(required_argument, named_argument=False, *args, **kwargs):
        return [required_argument, named_argument, args, kwargs]

    assert hug.test.cli(takes_all_the_things, 'hi!') == ['hi!', False, (), {}]
    assert hug.test.cli(takes_all_the_things, 'hi!', named_argument='there') == ['hi!', 'there', (), {}]
    assert hug.test.cli(takes_all_the_things, 'hi!', 'extra', '--arguments', 'can', '--happen', '--all', 'the', 'tim') \
                             == ['hi!', False, ('extra', ), {'arguments': 'can', 'happen': True, 'all': ['the', 'tim']}]


def test_api_gets_extra_variables_without_kargs_or_kwargs(hug_api):
    @hug.get(api=hug_api)
    def ensure_params(request, response):
        return request.params

    assert hug.test.get(hug_api, 'ensure_params', params={'make': 'it'}).data == {'make': 'it'}
    assert hug.test.get(hug_api, 'ensure_params', hello='world').data == {'hello': 'world'}



from tinydb import TinyDB, Query
import hug
import hashlib
import logging
import os


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
db = TinyDB('db.json')

"""
  Helper Methods
"""


def hash_password(password, salt):
    """
    Securely hash a password using a provided salt
    :param password:
    :param salt:
    :return: Hex encoded SHA512 hash of provided password
    """
    password = str(password).encode('utf-8')
    salt = str(salt).encode('utf-8')
    return hashlib.sha512(password + salt).hexdigest()


def gen_api_key(username):
    """
    Create a random API key for a user
    :param username:
    :return: Hex encoded SHA512 random string
    """
    salt = str(os.urandom(64)).encode('utf-8')
    return hash_password(username, salt)


@hug.cli()
def authenticate_user(username, password):
    """
    Authenticate a username and password against our database
    :param username:
    :param password:
    :return: authenticated username
    """
    user_model = Query()
    user = db.search(user_model.username == username)[0]

    if not user:
        logger.warning("User %s not found", username)
        return False

    if user['password'] == hash_password(password, user.get('salt')):
        return user['username']

    return False


@hug.cli()
def authenticate_key(api_key):
    """
    Authenticate an API key against our database
    :param api_key:
    :return: authenticated username
    """
    user_model = Query()
    user = db.search(user_model.api_key == api_key)[0]
    if user:
        return user['username']
    return False

"""
  API Methods start here
"""

api_key_authentication = hug.authentication.api_key(authenticate_key)
basic_authentication = hug.authentication.basic(authenticate_user)


@hug.cli()
def add_user(username, password):
    """
    CLI Parameter to add a user to the database
    :param username:
    :param password:
    :return: JSON status output
    """

    user_model = Query()
    if db.search(user_model.username == username):
        return {
            'error': 'User {0} already exists'.format(username)
        }

    salt = hashlib.sha512(str(os.urandom(64)).encode('utf-8')).hexdigest()
    password = hash_password(password, salt)
    api_key = gen_api_key(username)

    user = {
        'username': username,
        'password': password,
        'salt': salt,
        'api_key': api_key
    }
    user_id = db.insert(user)

    return {
       'result': 'success',
       'eid': user_id,
       'user_created': user
    }


@hug.get('/api/get_api_key', requires=basic_authentication)
def get_token(authed_user: hug.directives.user):
    """
    Get Job details
    :param authed_user:
    :return:
    """
    user_model = Query()
    user = db.search(user_model.username == authed_user)[0]

    if user:
        out = {
            'user': user['username'],
            'api_key': user['api_key']
        }
    else:
        # this should never happen
        out = {
            'error': 'User {0} does not exist'.format(authed_user)
        }

    return out


# Same thing, but authenticating against an API key
@hug.get(('/api/job', '/api/job/{job_id}/'), requires=api_key_authentication)
def get_job_details(job_id):
    """
    Get Job details
    :param job_id:
    :return:
    """
    job = {
        'job_id': job_id,
        'details': 'Details go here'
    }

    return job


if __name__ == '__main__':
    add_user.interface.cli()



"""A simple streaming movie server example"""
import hug


@hug.get(output=hug.output_format.mp4_video)
def watch():
    """Watch an example movie, streamed directly to you from hug"""
    return 'movie.mp4'



"""First hug API (local, command-line, and HTTP access)"""
import hug


@hug.cli()
@hug.get(examples='name=Timothy&age=26')
@hug.local()
def happy_birthday(name: hug.types.text, age: hug.types.number, hug_timer=3):
    """Says happy birthday to a user"""
    return {'message': 'Happy {0} Birthday {1}!'.format(age, name),
            'took': float(hug_timer)}


if __name__ == '__main__':
    happy_birthday.interface.cli()



"""tests/test_middleware.py.

Tests the middleware integrated into Hug

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import hug
import pytest
from falcon.request import SimpleCookie
from hug.exceptions import SessionNotFound
from hug.middleware import LogMiddleware, SessionMiddleware
from hug.store import InMemoryStore

api = hug.API(__name__)

# Fix flake8 undefined names (F821)
__hug__ = __hug__  # noqa


def test_session_middleware():
    @hug.get()
    def count(request):
        session = request.context['session']
        counter = session.get('counter', 0) + 1
        session['counter'] = counter
        return counter

    def get_cookies(response):
        simple_cookie = SimpleCookie(response.headers_dict['set-cookie'])
        return {morsel.key: morsel.value for morsel in simple_cookie.values()}

    # Add middleware
    session_store = InMemoryStore()
    middleware = SessionMiddleware(session_store, cookie_name='test-sid')
    __hug__.http.add_middleware(middleware)

    # Get cookies from response
    response = hug.test.get(api, '/count')
    cookies = get_cookies(response)

    # Assert session cookie has been set and session exists in session store
    assert 'test-sid' in cookies
    sid = cookies['test-sid']
    assert session_store.exists(sid)
    assert session_store.get(sid) == {'counter': 1}

    # Assert session persists throughout the requests
    headers = {'Cookie': 'test-sid={}'.format(sid)}
    assert hug.test.get(api, '/count', headers=headers).data == 2
    assert session_store.get(sid) == {'counter': 2}

    # Assert a non-existing session cookie gets ignored
    headers = {'Cookie': 'test-sid=foobarfoo'}
    response = hug.test.get(api, '/count', headers=headers)
    cookies = get_cookies(response)
    assert response.data == 1
    assert not session_store.exists('foobarfoo')
    assert cookies['test-sid'] != 'foobarfoo'


def test_logging_middleware():
    output = []

    class Logger(object):
        def info(self, content):
            output.append(content)

    @hug.middleware_class()
    class CustomLogger(LogMiddleware):
        def __init__(self, logger=Logger()):
            super().__init__(logger=logger)

    @hug.get()
    def test(request):
        return 'data'

    hug.test.get(api, '/test')
    assert output[0] == 'Requested: GET /test None'
    assert len(output[1]) > 0



"""A simple post reading server example.

To test run this server with `hug -f post_body_example`
then run the following from ipython:
    import requests

    requests.post('http://localhost:8000/post_here', json={'one': 'two'}).json()

    This should return back the json data that you posted
"""
import hug


@hug.post()
def post_here(body):
    """This example shows how to read in post data w/ hug outside of its automatic param parsing"""
    return body



"""tests/test_types.py.

Tests the type validators included with hug

Copyright (C) 2016 Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import urllib
from datetime import datetime
from decimal import Decimal
from uuid import UUID

import hug
import pytest
from hug.exceptions import InvalidTypeData
from marshmallow import Schema, fields


def test_type():
    """Test to ensure the abstract Type object can't be used"""
    with pytest.raises(NotImplementedError):
        hug.types.Type()('value')


def test_number():
    """Tests that hug's number type correctly converts and validates input"""
    assert hug.types.number('1') == 1
    assert hug.types.number(1) == 1
    with pytest.raises(ValueError):
        hug.types.number('bacon')


def test_range():
    """Tests that hug's range type successfully handles ranges of numbers"""
    assert hug.types.in_range(1, 10)('1') == 1
    assert hug.types.in_range(1, 10)(1) == 1
    assert '1' in hug.types.in_range(1, 10).__doc__
    with pytest.raises(ValueError):
        hug.types.in_range(1, 10)('bacon')
    with pytest.raises(ValueError):
        hug.types.in_range(1, 10)('15')
    with pytest.raises(ValueError):
        hug.types.in_range(1, 10)(-34)


def test_less_than():
    """Tests that hug's less than type successfully limits the values passed in"""
    assert hug.types.less_than(10)('1') == 1
    assert hug.types.less_than(10)(1) == 1
    assert hug.types.less_than(10)(-10) == -10
    assert '10' in hug.types.less_than(10).__doc__
    with pytest.raises(ValueError):
        assert hug.types.less_than(10)(10)


def test_greater_than():
    """Tests that hug's greater than type succefully limis the values passed in"""
    assert hug.types.greater_than(10)('11') == 11
    assert hug.types.greater_than(10)(11) == 11
    assert hug.types.greater_than(10)(1000) == 1000
    assert '10' in hug.types.greater_than(10).__doc__
    with pytest.raises(ValueError):
        assert hug.types.greater_than(10)(9)


def test_multiple():
    """Tests that hug's multile type correctly forces values to come back as lists, but not lists of lists"""
    assert hug.types.multiple('value') == ['value']
    assert hug.types.multiple(['value1', 'value2']) == ['value1', 'value2']


def test_delimited_list():
    """Test to ensure hug's custom delimited list type function works as expected"""
    assert hug.types.delimited_list(',')('value1,value2') == ['value1', 'value2']
    assert hug.types.delimited_list(',')(['value1', 'value2']) == ['value1', 'value2']
    assert hug.types.delimited_list('|-|')('value1|-|value2|-|value3,value4') == ['value1', 'value2', 'value3,value4']
    assert ',' in hug.types.delimited_list(',').__doc__


def test_comma_separated_list():
    """Tests that hug's comma separated type correctly converts into a Python list"""
    assert hug.types.comma_separated_list('value') == ['value']
    assert hug.types.comma_separated_list('value1,value2') == ['value1', 'value2']


def test_float_number():
    """Tests to ensure the float type correctly allows floating point values"""
    assert hug.types.float_number('1.1') == 1.1
    assert hug.types.float_number('1') == float(1)
    assert hug.types.float_number(1.1) == 1.1
    with pytest.raises(ValueError):
        hug.types.float_number('bacon')


def test_decimal():
    """Tests to ensure the decimal type correctly allows decimal values"""
    assert hug.types.decimal('1.1') == Decimal('1.1')
    assert hug.types.decimal('1') == Decimal('1')
    assert hug.types.decimal(1.1) == Decimal(1.1)
    with pytest.raises(ValueError):
        hug.types.decimal('bacon')


def test_boolean():
    """Test to ensure the custom boolean type correctly supports boolean conversion"""
    assert hug.types.boolean('1')
    assert hug.types.boolean('T')
    assert not hug.types.boolean('')
    assert hug.types.boolean('False')
    assert not hug.types.boolean(False)


def test_mapping():
    """Test to ensure the mapping type works as expected"""
    mapping_type = hug.types.mapping({'n': None, 'l': [], 's': set()})
    assert mapping_type('n') is None
    assert mapping_type('l') == []
    assert mapping_type('s') == set()
    assert 'n' in mapping_type.__doc__
    with pytest.raises(KeyError):
        mapping_type('bacon')


def test_smart_boolean():
    """Test to ensure that the smart boolean type works as expected"""
    assert hug.types.smart_boolean('true')
    assert hug.types.smart_boolean('t')
    assert hug.types.smart_boolean('1')
    assert hug.types.smart_boolean(1)
    assert not hug.types.smart_boolean('')
    assert not hug.types.smart_boolean('false')
    assert not hug.types.smart_boolean('f')
    assert not hug.types.smart_boolean('0')
    assert not hug.types.smart_boolean(0)
    assert hug.types.smart_boolean(True)
    assert not hug.types.smart_boolean(None)
    assert not hug.types.smart_boolean(False)
    with pytest.raises(KeyError):
        hug.types.smart_boolean('bacon')


def test_text():
    """Tests that hug's text validator correctly handles basic values"""
    assert hug.types.text('1') == '1'
    assert hug.types.text(1) == '1'
    assert hug.types.text('text') == 'text'
    with pytest.raises(ValueError):
        hug.types.text(['one', 'two'])

def test_uuid():
    """Tests that hug's text validator correctly handles UUID values
       Examples were taken from https://docs.python.org/3/library/uuid.html"""

    assert hug.types.uuid('{12345678-1234-5678-1234-567812345678}') == UUID('12345678-1234-5678-1234-567812345678')
    assert hug.types.uuid('12345678-1234-5678-1234-567812345678') == UUID('12345678123456781234567812345678')
    assert hug.types.uuid('12345678123456781234567812345678') == UUID('12345678-1234-5678-1234-567812345678')
    assert hug.types.uuid('urn:uuid:12345678-1234-5678-1234-567812345678') == \
           UUID('12345678-1234-5678-1234-567812345678')

    with pytest.raises(ValueError):
        hug.types.uuid(1)

    with pytest.raises(ValueError):
        # Invalid HEX character
        hug.types.uuid('12345678-1234-5678-1234-56781234567G')

    with pytest.raises(ValueError):
        # One character added
        hug.types.uuid('12345678-1234-5678-1234-5678123456781')
    with pytest.raises(ValueError):
        # One character removed
        hug.types.uuid('12345678-1234-5678-1234-56781234567')



def test_length():
    """Tests that hug's length type successfully handles a length range"""
    assert hug.types.length(1, 10)('bacon') == 'bacon'
    assert hug.types.length(1, 10)(42) == '42'
    assert '42' in hug.types.length(1, 42).__doc__
    with pytest.raises(ValueError):
        hug.types.length(1, 10)('bacon is the greatest food known to man')
    with pytest.raises(ValueError):
        hug.types.length(1, 10)('')
    with pytest.raises(ValueError):
        hug.types.length(1, 10)('bacon is th')


def test_shorter_than():
    """Tests that hug's shorter than type successfully limits the values passed in"""
    assert hug.types.shorter_than(10)('hi there') == 'hi there'
    assert hug.types.shorter_than(10)(1) == '1'
    assert hug.types.shorter_than(10)('') == ''
    assert '10' in hug.types.shorter_than(10).__doc__
    with pytest.raises(ValueError):
        assert hug.types.shorter_than(10)('there is quite a bit of text here, in fact way more than allowed')


def test_longer_than():
    """Tests that hug's greater than type succefully limis the values passed in"""
    assert hug.types.longer_than(10)('quite a bit of text here should be') == 'quite a bit of text here should be'
    assert hug.types.longer_than(10)(12345678910) == '12345678910'
    assert hug.types.longer_than(10)(100123456789100) == '100123456789100'
    assert '10' in hug.types.longer_than(10).__doc__
    with pytest.raises(ValueError):
        assert hug.types.longer_than(10)('short')


def test_cut_off():
    """Test to ensure that hug's cut_off type works as expected"""
    assert hug.types.cut_off(10)('text') == 'text'
    assert hug.types.cut_off(10)(10) == '10'
    assert hug.types.cut_off(10)('some really long text') == 'some reall'
    assert '10' in hug.types.cut_off(10).__doc__


def test_inline_dictionary():
    """Tests that inline dictionary values are correctly handled"""
    assert hug.types.inline_dictionary('1:2') == {'1': '2'}
    assert hug.types.inline_dictionary('1:2|3:4') == {'1': '2', '3': '4'}
    with pytest.raises(ValueError):
        hug.types.inline_dictionary('1')


def test_one_of():
    """Tests that hug allows limiting a value to one of a list of values"""
    assert hug.types.one_of(('bacon', 'sausage', 'pancakes'))('bacon') == 'bacon'
    assert hug.types.one_of(['bacon', 'sausage', 'pancakes'])('sausage') == 'sausage'
    assert hug.types.one_of({'bacon', 'sausage', 'pancakes'})('pancakes') == 'pancakes'
    assert 'bacon' in hug.types.one_of({'bacon', 'sausage', 'pancakes'}).__doc__
    with pytest.raises(KeyError):
        hug.types.one_of({'bacon', 'sausage', 'pancakes'})('syrup')


def test_accept():
    """Tests to ensure the accept type wrapper works as expected"""
    custom_converter = lambda value: value + " converted"
    custom_type = hug.types.accept(custom_converter, 'A string Value')
    with pytest.raises(TypeError):
        custom_type(1)


def test_accept_custom_exception_text():
    """Tests to ensure it's easy to custom the exception text using the accept wrapper"""
    custom_converter = lambda value: value + " converted"
    custom_type = hug.types.accept(custom_converter, 'A string Value', 'Error occurred')
    assert custom_type('bacon') == 'bacon converted'
    with pytest.raises(ValueError):
        custom_type(1)


def test_accept_custom_exception_handlers():
    """Tests to ensure it's easy to custom the exception text using the accept wrapper"""
    custom_converter = lambda value: (str(int(value)) if value else value) + " converted"
    custom_type = hug.types.accept(custom_converter, 'A string Value', exception_handlers={TypeError: '0 provided'})
    assert custom_type('1') == '1 converted'
    with pytest.raises(ValueError):
        custom_type('bacon')
    with pytest.raises(ValueError):
        custom_type(0)

    custom_type = hug.types.accept(custom_converter, 'A string Value', exception_handlers={TypeError: KeyError})
    with pytest.raises(KeyError):
        custom_type(0)


def test_json():
    """Test to ensure that the json type correctly handles url encoded json, as well as direct json"""
    assert hug.types.json({'this': 'works'}) == {'this': 'works'}
    assert hug.types.json(json.dumps({'this': 'works'})) == {'this': 'works'}
    with pytest.raises(ValueError):
        hug.types.json('Invalid JSON')


def test_multi():
    """Test to ensure that the multi type correctly handles a variety of value types"""
    multi_type = hug.types.multi(hug.types.json, hug.types.smart_boolean)
    assert multi_type({'this': 'works'}) == {'this': 'works'}
    assert multi_type(json.dumps({'this': 'works'})) == {'this': 'works'}
    assert multi_type('t')
    with pytest.raises(ValueError):
        multi_type('Bacon!')


def test_chain():
    """Test to ensure that chaining together multiple types works as expected"""
    chain_type = hug.types.Chain(hug.types.text, hug.types.LongerThan(10))
    assert chain_type(12345678901) == "12345678901"
    with pytest.raises(ValueError):
        chain_type(1)


def test_nullable():
    """Test the concept of a nullable type"""
    nullable_type = hug.types.Nullable(hug.types.text, hug.types.LongerThan(10))
    assert nullable_type(12345678901) == "12345678901"
    assert nullable_type(None) is None
    with pytest.raises(ValueError):
        nullable_type(1)


def test_schema_type():
    """Test hug's complex schema types"""
    class User(hug.types.Schema):
        username = hug.types.text
        password = hug.types.Chain(hug.types.text, hug.types.LongerThan(10))
    user_one = User({"username": "brandon", "password": "password123"})
    user_two = User(user_one)
    with pytest.raises(ValueError):
        user_three = User({"username": "brandon", "password": "123"})
    user_three = User({"username": "brandon", "password": "123"}, force=True)
    with pytest.raises(AttributeError):
        del user_one.username
    assert "username" in User.__slots__
    assert "_username" in User.__slots__
    assert user_one._username == "brandon"
    assert user_two == user_one
    assert user_three._username == "brandon"
    assert user_one.username == "brandon"
    assert user_two.username == "brandon"
    assert user_three.username == "brandon"
    assert user_one.password == "password123"
    with pytest.raises(ValueError):
        user_one.password = "test"
    assert user_one.password == "password123"


def test_marshmallow_schema():
    """Test hug's marshmallow schema support"""
    class UserSchema(Schema):
        name = fields.Str()

    schema_type = hug.types.MarshmallowSchema(UserSchema())
    assert schema_type({"name": "test"}) == {"name": "test"}
    assert schema_type("""{"name": "test"}""") == {"name": "test"}
    assert schema_type.__doc__ == 'UserSchema'
    with pytest.raises(InvalidTypeData):
        schema_type({"name": 1})


def test_create_type():
    """Test hug's new type creation decorator works as expected"""
    @hug.type(extend=hug.types.text, exception_handlers={TypeError: ValueError, LookupError: 'Hi!'},
              error_text='Invalid')
    def prefixed_string(value):
        if value == 'hi':
            raise TypeError('Repeat of prefix')
        elif value == 'bye':
            raise LookupError('Never say goodbye!')
        elif value == '1+1':
            raise ArithmeticError('Testing different error types')
        return 'hi-' + value

    assert prefixed_string('there') == 'hi-there'
    with pytest.raises(ValueError):
        prefixed_string([])
    with pytest.raises(ValueError):
        prefixed_string('hi')
    with pytest.raises(ValueError):
        prefixed_string('bye')

    @hug.type(extend=hug.types.text, exception_handlers={TypeError: ValueError})
    def prefixed_string(value):
        if value == '1+1':
            raise ArithmeticError('Testing different error types')
        return 'hi-' + value

    with pytest.raises(ArithmeticError):
        prefixed_string('1+1')

    @hug.type(extend=hug.types.text)
    def prefixed_string(value):
        return 'hi-' + value

    assert prefixed_string('there') == 'hi-there'

    @hug.type(extend=hug.types.one_of)
    def numbered(value):
        return int(value)

    assert numbered(['1', '2', '3'])('1') == 1



"""tests/test_documentation.py.

Tests the documentation generation capibilities integrated into Hug

Copyright (C) 2016 Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import json

import hug
from falcon import Request
from falcon.testing import StartResponseMock, create_environ

api = hug.API(__name__)


def test_basic_documentation():
    """Ensure creating and then documenting APIs with Hug works as intuitively as expected"""
    @hug.get()
    def hello_world():
        """Returns hello world"""
        return "Hello World!"

    @hug.post()
    def echo(text):
        """Returns back whatever data it is given in the text parameter"""
        return text

    @hug.post('/happy_birthday', examples="name=HUG&age=1")
    def birthday(name, age: hug.types.number=1):
        """Says happy birthday to a user"""
        return "Happy {age} Birthday {name}!".format(**locals())

    @hug.post()
    def noop(request, response):
        """Performs no action"""
        pass

    @hug.get()
    def string_docs(data: 'Takes data', ignore_directive: hug.directives.Timer) -> 'Returns data':
        """Annotations defined with strings should be documentation only"""
        pass

    @hug.get(private=True)
    def private():
        """Hidden from documentation"""
        pass

    documentation = api.http.documentation()
    assert 'test_documentation' in documentation['overview']

    assert '/hello_world' in documentation['handlers']
    assert '/echo' in documentation['handlers']
    assert '/happy_birthday' in documentation['handlers']
    assert not '/birthday' in documentation['handlers']
    assert '/noop' in documentation['handlers']
    assert '/string_docs' in documentation['handlers']
    assert not '/private' in documentation['handlers']

    assert documentation['handlers']['/hello_world']['GET']['usage'] == "Returns hello world"
    assert documentation['handlers']['/hello_world']['GET']['examples'] == ["/hello_world"]
    assert documentation['handlers']['/hello_world']['GET']['outputs']['content_type'] == "application/json"
    assert not 'inputs' in documentation['handlers']['/hello_world']['GET']

    assert 'text' in documentation['handlers']['/echo']['POST']['inputs']['text']['type']
    assert not 'default' in documentation['handlers']['/echo']['POST']['inputs']['text']

    assert 'number' in documentation['handlers']['/happy_birthday']['POST']['inputs']['age']['type']
    assert documentation['handlers']['/happy_birthday']['POST']['inputs']['age']['default'] == 1

    assert not 'inputs' in documentation['handlers']['/noop']['POST']

    assert documentation['handlers']['/string_docs']['GET']['inputs']['data']['type'] == 'Takes data'
    assert documentation['handlers']['/string_docs']['GET']['outputs']['type'] == 'Returns data'
    assert not 'ignore_directive' in documentation['handlers']['/string_docs']['GET']['inputs']

    @hug.post(versions=1)  # noqa
    def echo(text):
        """V1 Docs"""
        return 'V1'

    @hug.post(versions=2)  # noqa
    def echo(text):
        """V1 Docs"""
        return 'V2'

    @hug.post(versions=2)
    def test(text):
        """V1 Docs"""
        return True

    @hug.get(requires=test)
    def unversioned():
        return 'Hello'

    @hug.get(versions=False)
    def noversions():
        pass

    @hug.extend_api('/fake', base_url='/api')
    def extend_with():
        import tests.module_fake_simple
        return (tests.module_fake_simple, )

    versioned_doc = api.http.documentation()
    assert 'versions' in versioned_doc
    assert 1 in versioned_doc['versions']
    assert 2 in versioned_doc['versions']
    assert False not in versioned_doc['versions']
    assert '/unversioned' in versioned_doc['handlers']
    assert '/echo' in versioned_doc['handlers']
    assert '/test' in versioned_doc['handlers']

    specific_version_doc = api.http.documentation(api_version=1)
    assert 'versions' in specific_version_doc
    assert '/echo' in specific_version_doc['handlers']
    assert '/unversioned' in specific_version_doc['handlers']
    assert specific_version_doc['handlers']['/unversioned']['GET']['requires'] == ['V1 Docs']
    assert '/test' not in specific_version_doc['handlers']

    specific_base_doc = api.http.documentation(base_url='/api')
    assert '/echo' not in specific_base_doc['handlers']
    assert '/fake/made_up_hello' in specific_base_doc['handlers']

    handler = api.http.documentation_404()
    response = StartResponseMock()
    handler(Request(create_environ(path='v1/doc')), response)
    documentation = json.loads(response.data.decode('utf8'))['documentation']
    assert 'versions' in documentation
    assert '/echo' in documentation['handlers']
    assert '/test' not in documentation['handlers']



'''A basic example of authentication requests within a hug API'''
import hug
import jwt

# Several authenticators are included in hug/authentication.py. These functions
# accept a verify_user function, which can be either an included function (such
# as the basic username/password function demonstrated below), or logic of your
# own. Verification functions return an object to store in the request context
# on successful authentication. Naturally, this is a trivial demo, and a much
# more robust verification function is recommended. This is for strictly
# illustrative purposes.
authentication = hug.authentication.basic(hug.authentication.verify('User1', 'mypassword'))


@hug.get('/public')
def public_api_call():
    return "Needs no authentication"


# Note that the logged in user can be accessed via a built-in directive.
# Directives can provide computed input parameters via an abstraction
# layer so as not to clutter your API functions with access to the raw
# request object.
@hug.get('/authenticated', requires=authentication)
def basic_auth_api_call(user: hug.directives.user):
    return 'Successfully authenticated with user: {0}'.format(user)


# Here is a slightly less trivial example of how authentication might
# look in an API that uses keys.

# First, the user object stored in the context need not be a string,
# but can be any Python object.
class APIUser(object):
    """A minimal example of a rich User object"""

    def __init__(self, user_id, api_key):
        self.user_id = user_id
        self.api_key = api_key


def api_key_verify(api_key):
    magic_key = '5F00832B-DE24-4CAF-9638-C10D1C642C6C'  # Obviously, this would hit your database
    if api_key == magic_key:
        # Success!
        return APIUser('user_foo', api_key)
    else:
        # Invalid key
        return None


api_key_authentication = hug.authentication.api_key(api_key_verify)


@hug.get('/key_authenticated', requires=api_key_authentication)  # noqa
def basic_auth_api_call(user: hug.directives.user):
    return 'Successfully authenticated with user: {0}'.format(user.user_id)


def token_verify(token):
    secret_key = 'super-secret-key-please-change'
    try:
        return jwt.decode(token, secret_key, algorithm='HS256')
    except jwt.DecodeError:
        return False


token_key_authentication = hug.authentication.token(token_verify)


@hug.get('/token_authenticated', requires=token_key_authentication)  # noqa
def token_auth_call(user: hug.directives.user):
    return 'You are user: {0} with data {1}'.format(user['user'], user['data'])


@hug.post('/token_generation')  # noqa
def token_gen_call(username, password):
    """Authenticate and return a token"""
    secret_key = 'super-secret-key-please-change'
    mockusername = 'User2'
    mockpassword = 'Mypassword'
    if mockpassword == password and mockusername == username: # This is an example. Don't do that.
        return {"token" : jwt.encode({'user': username, 'data': 'mydata'}, secret_key, algorithm='HS256')}
    return 'Invalid username and/or password for user: {0}'.format(username)



"""An example of using a middleware to require HTTPS connections.
    requires https://github.com/falconry/falcon-require-https to be installed via
    pip install falcon-require-https
"""
import hug
from falcon_require_https import RequireHTTPS

hug.API(__name__).http.add_middleware(RequireHTTPS())


@hug.get()
def my_endpoint():
    return 'Success!'



#!/usr/bin/env python3
import tornado.ioloop
import tornado.web
import json


class TextHandler(tornado.web.RequestHandler):
    def get(self):
        self.write('Hello, world!')


application = tornado.web.Application([
    (r"/text", TextHandler),
])

if __name__ == "__main__":
    application.listen(8000)
    tornado.ioloop.IOLoop.current().start()



import hug
from PIL import Image, ImageDraw


@hug.get('/image.png', output=hug.output_format.png_image)
def create_image():
    image = Image.new('RGB', (100, 50)) # create the image
    ImageDraw.Draw(image).text((10, 10), 'Hello World!', fill=(255, 0, 0))
    return image



"""A basic (single function) API written using Hug"""
import hug


@hug.get('/happy_birthday', examples="name=HUG&age=1")
def happy_birthday(name, age: hug.types.number):
    """Says happy birthday to a user"""
    return "Happy {age} Birthday {name}!".format(**locals())



"""Example API using marshmallow fields as type annotations.


Requires marshmallow and dateutil.

    $ pip install marshmallow python-dateutil


To run the example:

    $ hug -f examples/marshmallow_example.py

Example requests using HTTPie:

    $ http :8000/dateadd value==1973-04-10 addend==63
    $ http :8000/dateadd value==2015-03-20 addend==525600 unit==minutes
"""
import datetime as dt

import hug
from marshmallow import fields
from marshmallow.validate import Range, OneOf


@hug.get('/dateadd', examples="value=1973-04-10&addend=63")
def dateadd(value: fields.DateTime(),
            addend: fields.Int(validate=Range(min=1)),
            unit: fields.Str(validate=OneOf(['minutes', 'days']))='days'):
    """Add a value to a date."""
    value = value or dt.datetime.utcnow()
    if unit == 'minutes':
        delta = dt.timedelta(minutes=addend)
    else:
        delta = dt.timedelta(days=addend)
    result = value + delta
    return {'result': result}



"""Fake HUG API module usable for testing importation of modules"""
import hug


class FakeException(BaseException):
    pass


@hug.directive(apply_globally=False)
def my_directive(default=None, **kwargs):
    """for testing"""
    return default


@hug.default_input_format('application/made-up')
def made_up_formatter(data):
    """for testing"""
    return data


@hug.default_output_format()
def output_formatter(data):
    """for testing"""
    return hug.output_format.json(data)


@hug.get()
def made_up_api(hug_my_directive=True):
    """for testing"""
    return hug_my_directive


@hug.directive(apply_globally=True)
def my_directive_global(default=None, **kwargs):
    """for testing"""
    return default


@hug.default_input_format('application/made-up', apply_globally=True)
def made_up_formatter_global(data):
    """for testing"""
    return data


@hug.default_output_format(apply_globally=True)
def output_formatter_global(data):
    """for testing"""
    return hug.output_format.json(data)


@hug.request_middleware()
def handle_request(request, response):
    """for testing"""
    return


@hug.startup()
def on_startup(api):
    """for testing"""
    return


@hug.static()
def static():
    """for testing"""
    return ('', )


@hug.sink('/all')
def sink(path):
    """for testing"""
    return path


@hug.exception(FakeException)
def handle_exception(exception):
    """Handles the provided exception for testing"""
    return True


@hug.not_found()
def not_found_handler():
    """for testing"""
    return True



"""First API, local access only"""
import hug


@hug.local()
def happy_birthday(name: hug.types.text, age: hug.types.number, hug_timer=3):
    """Says happy birthday to a user"""
    return {'message': 'Happy {0} Birthday {1}!'.format(age, name),
            'took': float(hug_timer)}



"""hug/test.py.

Tests the support for asynchronous method using asyncio coroutines

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import asyncio

import hug

loop = asyncio.get_event_loop()
api = hug.API(__name__)


def test_basic_call_coroutine():
    """The most basic Happy-Path test for Hug APIs using async"""
    @hug.call()
    @asyncio.coroutine
    def hello_world():
        return "Hello World!"

    assert loop.run_until_complete(hello_world()) == "Hello World!"


def test_nested_basic_call_coroutine():
    """The most basic Happy-Path test for Hug APIs using async"""
    @hug.call()
    @asyncio.coroutine
    def hello_world():
        return asyncio.async(nested_hello_world())

    @hug.local()
    @asyncio.coroutine
    def nested_hello_world():
        return "Hello World!"

    assert loop.run_until_complete(hello_world()) == "Hello World!"


def test_basic_call_on_method_coroutine():
    """Test to ensure the most basic call still works if applied to a method"""
    class API(object):

        @hug.call()
        @asyncio.coroutine
        def hello_world(self=None):
            return "Hello World!"

    api_instance = API()
    assert api_instance.hello_world.interface.http
    assert loop.run_until_complete(api_instance.hello_world()) == "Hello World!"
    assert hug.test.get(api, '/hello_world').data == "Hello World!"


def test_basic_call_on_method_through_api_instance_coroutine():
    """Test to ensure the most basic call still works if applied to a method"""
    class API(object):

        def hello_world(self):
            return "Hello World!"

    api_instance = API()

    @hug.call()
    @asyncio.coroutine
    def hello_world():
        return api_instance.hello_world()

    assert api_instance.hello_world() == "Hello World!"
    assert hug.test.get(api, '/hello_world').data == "Hello World!"


def test_basic_call_on_method_registering_without_decorator_coroutine():
    """Test to ensure instance method calling via async works as expected"""
    class API(object):

        def __init__(self):
            hug.call()(self.hello_world_method)

        @asyncio.coroutine
        def hello_world_method(self):
            return "Hello World!"

    api_instance = API()

    assert loop.run_until_complete(api_instance.hello_world_method()) == "Hello World!"
    assert hug.test.get(api, '/hello_world_method').data == "Hello World!"



"""A basic example of using hug.use.Socket to return data from raw sockets"""
import hug
import socket
import struct
import time


http_socket = hug.use.Socket(connect_to=('www.google.com', 80), proto='tcp', pool=4, timeout=10.0)
ntp_service = hug.use.Socket(connect_to=('127.0.0.1', 123), proto='udp', pool=4, timeout=10.0)


EPOCH_START = 2208988800
@hug.get()
def get_time():
    """Get time from a locally running NTP server"""

    time_request = '\x1b' + 47 * '\0'
    now = struct.unpack("!12I", ntp_service.request(time_request, timeout=5.0).data.read())[10]
    return time.ctime(now - EPOCH_START)


@hug.get()
def reverse_http_proxy(length: int=100):
    """Simple reverse http proxy function that returns data/html from another http server (via sockets)
    only drawback is the peername is static, and currently does not support being changed.
    Example: curl localhost:8000/reverse_http_proxy?length=400"""

    http_request = """
GET / HTTP/1.0\r\n\r\n
Host: www.google.com\r\n\r\n
\r\n\r\n
"""
    return http_socket.request(http_request, timeout=5.0).data.read()[0:length]



import hug

API = hug.API('git')


@hug.object(name='git', version='1.0.0', api=API)
class GIT(object):
    """An example of command like calls via an Object"""

    @hug.object.cli
    def push(self, branch='master'):
        return 'Pushing {}'.format(branch)

    @hug.object.cli
    def pull(self, branch='master'):
        return 'Pulling {}'.format(branch)


if __name__ == '__main__':
    API.cli()



"""A basic cli client written with hug"""
import hug


@hug.cli(version="1.0.0")
def cli(name: 'The name', age: hug.types.number):
    """Says happy birthday to a user"""
    return "Happy {age} Birthday {name}!\n".format(**locals())


if __name__ == '__main__':
    cli.interface.cli()
